from vint.asset import get_asset_path
from vint.linting.config.config_file_source import ConfigFileSource


DEFAULT_CONFIG_PATH = get_asset_path('default_config.yaml')


class ConfigDefaultSource(ConfigFileSource):
    def get_file_path(self, env):
        return DEFAULT_CONFIG_PATH



class AbstractPolicy(object):
    def __init__(self):
        self.name = self.__class__.__name__
        self.description = None
        self.reference = None
        self.level = None


    def listen_node_types(self):
        """ Listening node type.
        is_valid will be called when a linter visit the listening node type.
        """
        return []


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid for the policy. """
        return True


    def create_violation_report(self, node, lint_context):
        """ Returns a violation report for the node. """
        return {
            'name': self.name,
            'level': self.level,
            'description': self.description,
            'reference': self.reference,
            'position': {
                'line': node['pos']['lnum'],
                'column': node['pos']['col'],
                'path': lint_context['path'],
            },
        }


    def get_violation_if_found(self, node, lint_context):
        """ Returns a violation if the node is invalid. """
        if self.is_valid(node, lint_context):
            return None

        return self.create_violation_report(node, lint_context)


    def get_policy_options(self, lint_context):
        policy_section = lint_context['config'].get('policies', {})
        return policy_section.get(self.name, {})



import re
from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy


@register_policy
class ProhibitEncodingOptionAfterScriptEncoding(AbstractPolicy):
    def __init__(self):
        super(ProhibitEncodingOptionAfterScriptEncoding, self).__init__()
        self.description = 'Set encoding before setting scriptencoding'
        self.reference = ':help :scriptencoding'
        self.level = Level.WARNING

        self.was_scriptencoding_found = False
        self.has_encoding_opt_after_scriptencoding = False


    def listen_node_types(self):
        return [NodeType.EXCMD]


    def is_valid(self, excmd_node, lint_context):
        """ Whether the specified node is valid.

        This policy prohibits encoding option after scriptencoding.
        """

        cmd_str = excmd_node['str']

        if re.match(r':*scripte', cmd_str):
            self.was_scriptencoding_found = True

        if re.match(r':*set? +enc', cmd_str) and self.was_scriptencoding_found:
            return False

        return True



import json
from pathlib import Path



class JSONFormatter(object):
    def __init__(self, env):
        pass


    def format_violations(self, violations):
        return json.dumps(self._normalize_violations(violations))


    def _normalize_violations(self, violations):
        line_number = lambda violation: violation['position']['line']
        sorted_violations = sorted(violations, key=line_number)

        normalized_violations = map(self._normalize_violation, sorted_violations)

        return list(normalized_violations)


    def _normalize_violation(self, violation):
        return {
            'file_path': str(Path(violation['position']['path'])),
            'line_number': violation['position']['line'],
            'column_number': violation['position']['column'],
            'severity': violation['level'].name.lower(),
            'description': violation['description'],
            'policy_name': violation['name'],
            'reference': violation['reference'],
        }



class TwoWayScopeReferenceAttacher(object):
    @classmethod
    def attach(cls, scope_tree):
        for child_scope in scope_tree['child_scopes']:
            child_scope['parent_scope'] = scope_tree
            cls.attach(child_scope)

        return scope_tree



from vint.linting.formatter.formatter import Formatter


class StatisticFormatter(Formatter):
    def format_violations(self, violations):
        violations_count = len(violations)

        output = super(StatisticFormatter, self).format_violations(violations) + '\n'
        return output + 'Total violations: {count}'.format(count=violations_count)



from functools import reduce
from vint.linting.config.config_source import ConfigSource



class ConfigContainer(ConfigSource):
    def __init__(self, *config_sources):
        self.config_sources = config_sources


    def get_config_dict(self):
        def extends_deeply(dict_to_extend, prior_dict):
            for key, value in prior_dict.items():
                if isinstance(value, dict):
                    if key in dict_to_extend:
                        extends_deeply(dict_to_extend[key], value)
                    else:
                        dict_to_extend[key] = value
                else:
                    dict_to_extend[key] = value

            return dict_to_extend


        config_dicts_ordered_by_prior_asc = [config_source.get_config_dict()
                                             for config_source in self.config_sources]

        return reduce(extends_deeply, config_dicts_ordered_by_prior_asc, {})



from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy.reference.googlevimscriptstyleguide import get_reference_source
from vint.linting.level import Level
from vint.ast.node_type import NodeType
from vint.linting.policy_registry import register_policy


PROHIBITED_COMMAND_PATTERNS = ('substitute',
                               '&',
                               '~',
                               'snomagic',
                               'smagic')


@register_policy
class ProhibitCommandWithUnintendedSideEffect(AbstractPolicy):
    def __init__(self):
        super(ProhibitCommandWithUnintendedSideEffect, self).__init__()
        self.level = Level.WARNING
        self.description = 'Do not use a command that has unintended side effects'
        self.reference = get_reference_source('DANGEROUS')


    def listen_node_types(self):
        return [
            NodeType.EXCMD,
        ]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid to the policy.
        This policy prohibit using `:s[ubstitute]` family.
        """

        command = node['ea']['cmd'].get('name', None)
        is_prohibited_command = any(pattern == command
                                    for pattern in PROHIBITED_COMMAND_PATTERNS)

        return not is_prohibited_command



# See :help option-summary
Abbreviations = {
    'acd': 'autochdir',
    'ai': 'autoindent',
    'akm': 'altkeymap',
    'al': 'aleph',
    'ambw': 'ambiwidth',
    'anti': 'antialias',
    'ar': 'autoread',
    'arab': 'arabic',
    'ari': 'allowrevins',
    'arshape': 'arabicshape',
    'aw': 'autowrite',
    'awa': 'autowriteall',
    'bdir': 'backupdir',
    'bdlay': 'balloondelay',
    'beval': 'ballooneval',
    'bex': 'backupext',
    'bexpr': 'balloonexpr',
    'bg': 'background',
    'bh': 'bufhidden',
    'bin': 'binary',
    'biosk': 'bioskey',
    'bk': 'backup',
    'bkc': 'backupcopy',
    'bl': 'buflisted',
    'bri': 'breakindent',
    'briopt': 'breakindentopt',
    'brk': 'breakat',
    'bs': 'backspace',
    'bsdir': 'browsedir',
    'bsk': 'backupskip',
    'bt': 'buftype',
    'cb': 'clipboard',
    'cc': 'colorcolumn',
    'ccv': 'charconvert',
    'cd': 'cdpath',
    'cf': 'confirm',
    'cfu': 'completefunc',
    'ch': 'cmdheight',
    'ci': 'copyindent',
    'cin': 'cindent',
    'cink': 'cinkeys',
    'cino': 'cinoptions',
    'cinw': 'cinwords',
    'cm': 'cryptmethod',
    'cmp': 'casemap',
    'cms': 'commentstring',
    'co': 'columns',
    'cocu': 'concealcursor',
    'cole': 'conceallevel',
    'com': 'comments',
    'consk': 'conskey',
    'cot': 'completeopt',
    'cp': 'compatible',
    'cpo': 'cpoptions',
    'cpt': 'complete',
    'crb': 'cursorbind',
    'cspc': 'cscopepathcomp',
    'csprg': 'cscopeprg',
    'csqf': 'cscopequickfix',
    'csre': 'cscoperelative',
    'cst': 'cscopetag',
    'csto': 'cscopetagorder',
    'csverb': 'cscopeverbose',
    'cuc': 'cursorcolumn',
    'cul': 'cursorline',
    'cwh': 'cmdwinheight',
    'deco': 'delcombine',
    'def': 'define',
    'dex': 'diffexpr',
    'dg': 'digraph',
    'dict': 'dictionary',
    'dip': 'diffopt',
    'dir': 'directory',
    'dy': 'display',
    'ea': 'equalalways',
    'ead': 'eadirection',
    'eb': 'errorbells',
    'ed': 'edcompatible',
    'ef': 'errorfile',
    'efm': 'errorformat',
    'ei': 'eventignore',
    'ek': 'esckeys',
    'enc': 'encoding',
    'eol': 'endofline',
    'ep': 'equalprg',
    'et': 'expandtab',
    'ex': 'exrc',
    'fcl': 'foldclose',
    'fcs': 'fillchars',
    'fdc': 'foldcolumn',
    'fde': 'foldexpr',
    'fdi': 'foldignore',
    'fdl': 'foldlevel',
    'fdls': 'foldlevelstart',
    'fdm': 'foldmethod',
    'fdn': 'foldnestmax',
    'fdo': 'foldopen',
    'fdt': 'foldtext',
    'fen': 'foldenable',
    'fenc': 'fileencoding',
    'fencs': 'fileencodings',
    'fex': 'formatexpr',
    'ff': 'fileformat',
    'ffs': 'fileformats',
    'fic': 'fileignorecase',
    'fk': 'fkmap',
    'flp': 'formatlistpat',
    'fml': 'foldminlines',
    'fmr': 'foldmarker',
    'fo': 'formatoptions',
    'fp': 'formatprg',
    'fs': 'fsync',
    'ft': 'filetype',
    'gcr': 'guicursor',
    'gd': 'gdefault',
    'gfm': 'grepformat',
    'gfn': 'guifont',
    'gfs': 'guifontset',
    'gfw': 'guifontwide',
    'ghr': 'guiheadroom',
    'go': 'guioptions',
    'gp': 'grepprg',
    'gtl': 'guitablabel',
    'gtt': 'guitabtooltip',
    'hf': 'helpfile',
    'hh': 'helpheight',
    'hi': 'history',
    'hid': 'hidden',
    'hk': 'hkmap',
    'hkp': 'hkmapp',
    'hl': 'highlight',
    'hlg': 'helplang',
    'hls': 'hlsearch',
    'ic': 'ignorecase',
    'im': 'insertmode',
    'imaf': 'imactivatefunc',
    'imak': 'imactivatekey',
    'imc': 'imcmdline',
    'imd': 'imdisable',
    'imi': 'iminsert',
    'ims': 'imsearch',
    'imsf': 'imstatusfunc',
    'inc': 'include',
    'inde': 'indentexpr',
    'indk': 'indentkeys',
    'inex': 'includeexpr',
    'inf': 'infercase',
    'is': 'incsearch',
    'isf': 'isfname',
    'isi': 'isident',
    'isk': 'iskeyword',
    'isp': 'isprint',
    'js': 'joinspaces',
    'km': 'keymodel',
    'kmp': 'keymap',
    'kp': 'keywordprg',
    'lbr': 'linebreak',
    'lcs': 'listchars',
    'lm': 'langmenu',
    'lmap': 'langmap',
    'lpl': 'loadplugins',
    'ls': 'laststatus',
    'lsp': 'linespace',
    'lw': 'lispwords',
    'lz': 'lazyredraw',
    'ma': 'modifiable',
    'mat': 'matchtime',
    'mco': 'maxcombine',
    'mef': 'makeef',
    'mfd': 'maxfuncdepth',
    'mh': 'mousehide',
    'mis': 'menuitems',
    'ml': 'modeline',
    'mls': 'modelines',
    'mm': 'maxmem',
    'mmd': 'maxmapdepth',
    'mmp': 'maxmempattern',
    'mmt': 'maxmemtot',
    'mod': 'modified',
    'mousef': 'mousefocus',
    'mousem': 'mousemodel',
    'mouses': 'mouseshape',
    'mouset': 'mousetime',
    'mp': 'makeprg',
    'mps': 'matchpairs',
    'msm': 'mkspellmem',
    'mzq': 'mzquantum',
    'nf': 'nrformats',
    'nu': 'number',
    'nuw': 'numberwidth',
    'odev': 'opendevice',
    'oft': 'osfiletype',
    'ofu': 'omnifunc',
    'opfunc': 'operatorfunc',
    'pa': 'path',
    'para': 'paragraphs',
    'pdev': 'printdevice',
    'penc': 'printencoding',
    'pex': 'patchexpr',
    'pexpr': 'printexpr',
    'pfn': 'printfont',
    'ph': 'pumheight',
    'pheader': 'printheader',
    'pi': 'preserveindent',
    'pm': 'patchmode',
    'pmbcs': 'printmbcharset',
    'pmbfn': 'printmbfont',
    'popt': 'printoptions',
    'pt': 'pastetoggle',
    'pvh': 'previewheight',
    'pvw': 'previewwindow',
    'qe': 'quoteescape',
    'rdt': 'redrawtime',
    're': 'regexpengine',
    'ri': 'revins',
    'rl': 'rightleft',
    'rlc': 'rightleftcmd',
    'rnu': 'relativenumber',
    'ro': 'readonly',
    'rop': 'renderoptions',
    'rs': 'restorescreen',
    'rtp': 'runtimepath',
    'ru': 'ruler',
    'ruf': 'rulerformat',
    'sb': 'splitbelow',
    'sbo': 'scrollopt',
    'sbr': 'showbreak',
    'sc': 'showcmd',
    'scb': 'scrollbind',
    'scr': 'scroll',
    'scs': 'smartcase',
    'sect': 'sections',
    'sel': 'selection',
    'sft': 'showfulltag',
    'sh': 'shell',
    'shcf': 'shellcmdflag',
    'shm': 'shortmess',
    'shq': 'shellquote',
    'si': 'smartindent',
    'siso': 'sidescrolloff',
    'sj': 'scrolljump',
    'slm': 'selectmode',
    'sm': 'showmatch',
    'smc': 'synmaxcol',
    'smd': 'showmode',
    'sn': 'shortname',
    'so': 'scrolloff',
    'sol': 'startofline',
    'sp': 'shellpipe',
    'spc': 'spellcapcheck',
    'spf': 'spellfile',
    'spl': 'spelllang',
    'spr': 'splitright',
    'sps': 'spellsuggest',
    'sr': 'shiftround',
    'srr': 'shellredir',
    'ss': 'sidescroll',
    'ssl': 'shellslash',
    'ssop': 'sessionoptions',
    'st': 'shelltype',
    'sta': 'smarttab',
    'stal': 'showtabline',
    'stl': 'statusline',
    'stmp': 'shelltemp',
    'sts': 'softtabstop',
    'su': 'suffixes',
    'sua': 'suffixesadd',
    'sw': 'shiftwidth',
    'swb': 'switchbuf',
    'swf': 'swapfile',
    'sws': 'swapsync',
    'sxe': 'shellxescape',
    'sxq': 'shellxquote',
    'syn': 'syntax',
    'ta': 'textauto',
    'tag': 'tags',
    'tal': 'tabline',
    'tb': 'toolbar',
    'tbi': 'ttybuiltin',
    'tbidi': 'termbidi',
    'tbis': 'toolbariconsize',
    'tbs': 'tagbsearch',
    'tenc': 'termencoding',
    'tf': 'ttyfast',
    'tgst': 'tagstack',
    'tl': 'taglength',
    'tm': 'timeoutlen',
    'to': 'timeout',
    'top': 'tildeop',
    'tpm': 'tabpagemax',
    'tr': 'tagrelative',
    'ts': 'tabstop',
    'tsl': 'ttyscroll',
    'tsr': 'thesaurus',
    'ttm': 'ttimeoutlen',
    'tty': 'ttytype',
    'ttym': 'ttymouse',
    'tw': 'textwidth',
    'tx': 'textmode',
    'uc': 'updatecount',
    'udf': 'undofile',
    'udir': 'undodir',
    'ul': 'undolevels',
    'ur': 'undoreload',
    'ut': 'updatetime',
    'vb': 'visualbell',
    'vbs': 'verbose',
    'vdir': 'viewdir',
    've': 'virtualedit',
    'vfile': 'verbosefile',
    'vi': 'viminfo',
    'vop': 'viewoptions',
    'wa': 'writeany',
    'wak': 'winaltkeys',
    'wb': 'writebackup',
    'wc': 'wildchar',
    'wcm': 'wildcharm',
    'wd': 'writedelay',
    'wfh': 'winfixheight',
    'wfw': 'winfixwidth',
    'wh': 'winheight',
    'wi': 'window',
    'wic': 'wildignorecase',
    'wig': 'wildignore',
    'wim': 'wildmode',
    'wiv': 'weirdinvert',
    'wiw': 'winwidth',
    'wm': 'wrapmargin',
    'wmh': 'winminheight',
    'wmnu': 'wildmenu',
    'wmw': 'winminwidth',
    'wop': 'wildoptions',
    'ws': 'wrapscan',
    'ww': 'whichwrap',
}


AbbreviationsIncludingInvertPrefix = {}

for abbr_name, full_name in Abbreviations.items():
    AbbreviationsIncludingInvertPrefix[abbr_name] = full_name

    # See :help E518
    AbbreviationsIncludingInvertPrefix['no' + abbr_name] = 'no' + full_name

    # See :help :set-inv
    AbbreviationsIncludingInvertPrefix['inv' + abbr_name] = 'inv' + full_name



import re
from vint._bundles import vimlparser
from vint.ast.traversing import traverse
from vint.encodings.decoder import Decoder
from vint.encodings.decoding_strategy import default_decoding_strategy


class Parser(object):
    def __init__(self, plugins=None, enable_neovim=False):
        """ Initialize Parser with the specified plugins.
        The plugins can add attributes to the AST.
        """
        self.plugins = plugins.values() if plugins else []
        self._enable_neovim = enable_neovim


    def parse(self, string):
        """ Parse vim script string and return the AST. """
        lines = string.split('\n')

        reader = vimlparser.StringReader(lines)
        parser = vimlparser.VimLParser(self._enable_neovim)
        ast = parser.parse(reader)

        # TOPLEVEL does not have a pos, but we need pos for all nodes
        ast['pos'] = {'col': 1, 'i': 0, 'lnum': 1}

        for plugin in self.plugins:
            plugin.process(ast)

        return ast


    def parse_file(self, file_path):
        """ Parse vim script file and return the AST. """
        decoder = Decoder(default_decoding_strategy)
        decoded = decoder.read(file_path)
        decoded_and_lf_normalized = decoded.replace('\r\n', '\n')

        return self.parse(decoded_and_lf_normalized)


    def parse_redir(self, redir_cmd):
        """ Parse a command :redir content. """
        redir_cmd_str = redir_cmd['str']

        matched = re.match(r'redir?!?\s*(=>>?\s*)(\S+)', redir_cmd_str)
        if matched:
            redir_cmd_op = matched.group(1)
            redir_cmd_body = matched.group(2)

            arg_pos = redir_cmd['ea']['argpos']

            # Position of the "redir_cmd_body"
            start_pos = {
                'col': arg_pos['col'] + len(redir_cmd_op),
                'i': arg_pos['i'] + len(redir_cmd_op),
                'lnum': arg_pos['lnum'],
            }

            # NOTE: This is a hack to parse variable node.
            raw_ast = self.parse('echo ' + redir_cmd_body)

            # We need the left node of ECHO node
            redir_cmd_ast = raw_ast['body'][0]['list'][0]

            def adjust_position(node):
                pos = node['pos']
                # Care 1-based index and the length of "echo ".
                pos['col'] += start_pos['col'] - 1 - 5

                # Care the length of "echo ".
                pos['i'] += start_pos['i'] - 5

                # Care 1-based index
                pos['lnum'] += start_pos['lnum'] - 1

            traverse(redir_cmd_ast, on_enter=adjust_position)

            return redir_cmd_ast

        return None


    def parse_string_expr(self, string_expr_node):
        """ Parse a string node content. """
        string_expr_node_value = string_expr_node['value']
        string_expr_str = string_expr_node_value[1:-1]

        # Care escaped string literals
        if string_expr_node_value[0] == "'":
            string_expr_str = string_expr_str.replace("''", "'")
        else:
            string_expr_str = string_expr_str.replace('\\"', '"')

        # NOTE: This is a hack to parse expr1. See :help expr1
        raw_ast = self.parse('echo ' + string_expr_str)

        # We need the left node of ECHO node
        parsed_string_expr_nodes = raw_ast['body'][0]['list']

        start_pos = string_expr_node['pos']

        def adjust_position(node):
            pos = node['pos']

            # Care 1-based index and the length of "echo ".
            pos['col'] += start_pos['col'] - 1 - 5

            # Care the length of "echo ".
            pos['i'] += start_pos['i'] - 5

            # Care 1-based index
            pos['lnum'] += start_pos['lnum'] - 1

        for parsed_string_expr_node in parsed_string_expr_nodes:
            traverse(parsed_string_expr_node, on_enter=adjust_position)

        return parsed_string_expr_nodes



class ConfigSource(object):
    class ConfigError(Exception):
        def __init__(self, msg):
            self.msg = msg

        def __str__(self):
            return self.msg


    def get_config_dict(self):
        raise NotImplementedError()



import enum
from vint.ast.node_type import NodeType
from vint.ast.dictionary.builtins import (
    BuiltinVariablesCanHaveImplicitScope,
    BuiltinFunctions,
)
from vint.ast.plugin.scope_plugin.identifier_classifier import (
    is_dynamic_identifier,
    is_declarative_identifier,
    is_function_identifier,
    is_member_identifier,
    is_declarative_parameter,
    is_on_string_expression_context,
)


class ScopeVisibility(enum.Enum):
    """ 5 types scope visibility.
    We interest to analyze the variable scope by checking a single file.
    So, we do not interest to the strict visibility of the scopes that have a
    larger visibility than script local.
    """
    GLOBAL_LIKE = 0
    BUILTIN = 1
    SCRIPT_LOCAL = 2
    FUNCTION_LOCAL = 3
    UNANALYZABLE = 4
    INVALID = 5


class ExplicityOfScopeVisibility(enum.Enum):
    EXPLICIT = 0
    IMPLICIT = 1
    UNANALYZABLE = 2


FunctionDeclarationIdentifierScopePrefixToScopeVisibility = {
    'g:': ScopeVisibility.GLOBAL_LIKE,
    'b:': ScopeVisibility.INVALID,
    'w:': ScopeVisibility.INVALID,
    't:': ScopeVisibility.INVALID,
    's:': ScopeVisibility.SCRIPT_LOCAL,
    'l:': ScopeVisibility.INVALID,
    'a:': ScopeVisibility.INVALID,
    'v:': ScopeVisibility.INVALID,
}


VariableIdentifierScopePrefixToScopeVisibility = {
    'g:': ScopeVisibility.GLOBAL_LIKE,
    'b:': ScopeVisibility.GLOBAL_LIKE,
    'w:': ScopeVisibility.GLOBAL_LIKE,
    't:': ScopeVisibility.GLOBAL_LIKE,
    's:': ScopeVisibility.SCRIPT_LOCAL,
    'l:': ScopeVisibility.FUNCTION_LOCAL,
    'a:': ScopeVisibility.FUNCTION_LOCAL,
    'v:': ScopeVisibility.BUILTIN,
}


ImplicitScopeVisibilityToIdentifierScopePrefix = {
    ScopeVisibility.GLOBAL_LIKE: 'g:',
    ScopeVisibility.FUNCTION_LOCAL: 'l:',
    ScopeVisibility.BUILTIN: 'v:',
}


GlobalLikeScopeVisibilityNodeTypes = {
    NodeType.ENV: True,
    NodeType.OPTION: True,
    NodeType.REG: True,
}


IdentifierLikeNodeTypes = {
    NodeType.IDENTIFIER: True,
    NodeType.ENV: True,
    NodeType.OPTION: True,
    NodeType.REG: True,
}


def is_builtin_variable(id_node):
    """ Whether the specified node is a builtin identifier. """
    # Builtin variables are always IDENTIFIER.
    if NodeType(id_node['type']) is not NodeType.IDENTIFIER:
        return False

    id_value = id_node['value']

    if id_value.startswith('v:'):
        # It is an explicit builtin variable such as: "v:count", "v:char"
        # TODO: Add unknown builtin flag
        return True

    if is_builtin_function(id_node):
        return True

    if id_value in ['key', 'val']:
        # These builtin variable names are available on only map() or filter().
        return is_on_string_expression_context(id_node)

    # It is an implicit builtin variable such as: "count", "char"
    return id_value in BuiltinVariablesCanHaveImplicitScope


def is_builtin_function(id_node):
    """ Whether the specified node is a builtin function name identifier.
    The given identifier should be a child node of NodeType.CALL.
    """
    # Builtin functions are always IDENTIFIER.
    if NodeType(id_node['type']) is not NodeType.IDENTIFIER:
        return False

    id_value = id_node['value']

    if not is_function_identifier(id_node):
        return False

    # There are defference between a function identifier and variable
    # identifier:
    #
    #   let localtime = 0
    #   echo localtime " => 0
    #   echo localtime() " => 1420011455
    return id_value in BuiltinFunctions


def is_analyzable_identifier(node):
    """ Whether the specified node is an analyzable identifier.

    Node declarative-identifier-like is analyzable if it is not dynamic
    and not a member variable, because we can do static scope analysis.

    Analyzable cases:
      - let s:var = 0
      - function! Func()
      - echo s:var

    Unanalyzable cases:
      - let s:my_{var} = 0
      - function! dict.Func()
      - echo s:my_{var}
    """
    return not (is_dynamic_identifier(node) or is_member_identifier(node))


def is_analyzable_declarative_identifier(node):
    """ Whether the specified node is an analyzable declarative identifier.
    Node declarative-identifier-like is analyzable if it is not dynamic
    and not a member variable, because we can do static scope analysis.

    Analyzable cases:
      - let s:var = 0
      - function! Func()

    Unanalyzable cases:
      - let s:my_{var} = 0
      - function! dict.Func()
    """
    return is_declarative_identifier(node) and is_analyzable_identifier(node)


def detect_scope_visibility(node, context_scope):
    """ Returns a variable visibility hint by the specified node.
    The hint is a dict that has 2 attributes: "scope_visibility" and
    "is_implicit".
    """
    node_type = NodeType(node['type'])

    if not is_analyzable_identifier(node):
        return _create_identifier_visibility_hint(ScopeVisibility.UNANALYZABLE)

    if node_type is NodeType.IDENTIFIER:
        return _detect_identifier_scope_visibility(node, context_scope)

    if node_type in GlobalLikeScopeVisibilityNodeTypes:
        return _create_identifier_visibility_hint(ScopeVisibility.GLOBAL_LIKE)


def get_explicity_of_scope_visibility(node):
    """ Whether the node has a explicit scope visibility such as "a:".
    Returns True if the node is unanalyzable.
    """
    if not is_analyzable_identifier(node):
        return ExplicityOfScopeVisibility.UNANALYZABLE

    if is_declarative_parameter(node):
        # Declarative parameter can not have any explicit scope visibility
        # prefix, but we almost understand the identifier has an explicit scope
        # visibility that is function local.
        return ExplicityOfScopeVisibility.EXPLICIT

    if is_builtin_function(node):
        # Builtin functions can not have explicit scope visibility, but we
        # almost understand the identifier has an explicit scope visibility.
        return ExplicityOfScopeVisibility.EXPLICIT

    is_explicit = _get_explicit_scope_visibility(node) is not None

    return ExplicityOfScopeVisibility.EXPLICIT if is_explicit \
        else ExplicityOfScopeVisibility.IMPLICIT


def normalize_variable_name(node, context_scope):
    """ Returns normalized variable name.
    Normalizing means that variable names get explicit visibility by
    visibility prefix such as: "g:", "s:", ...

    Retunes None if the specified node is unanalyzable.
    A node is unanalyzable if:

    - the node is not identifier-like
    - the node is named dynamically
    """
    node_type = NodeType(node['type'])

    if not is_analyzable_identifier(node):
        return None

    if node_type is NodeType.IDENTIFIER:
        return _normalize_identifier_value(node, context_scope)

    # Nodes identifier-like without identifier is always normalized because
    # the nodes can not have a visibility prefix.
    if node_type in IdentifierLikeNodeTypes:
        return node['value']


def _normalize_identifier_value(id_node, context_scope):
    scope_visibility_hint = detect_scope_visibility(id_node, context_scope)

    if not scope_visibility_hint['is_implicit']:
        return id_node['value']

    scope_visibility = scope_visibility_hint['scope_visibility']

    if scope_visibility is ScopeVisibility.BUILTIN and is_function_identifier(id_node):
        # Builtin functions can not have explicit scope visibility.
        return id_node['value']

    scope_prefix = ImplicitScopeVisibilityToIdentifierScopePrefix[scope_visibility]

    return scope_prefix + id_node['value']


def _detect_identifier_scope_visibility(id_node, context_scope):
    explicit_scope_visibility = _get_explicit_scope_visibility(id_node)
    if explicit_scope_visibility is not None:
        return _create_identifier_visibility_hint(explicit_scope_visibility)

    is_implicit = get_explicity_of_scope_visibility(id_node) \
        is ExplicityOfScopeVisibility.IMPLICIT

    # Implicit scope variable will be resolved as a builtin variable if it
    # has a same name to Vim builtin variables.
    if is_builtin_variable(id_node):
        return _create_identifier_visibility_hint(ScopeVisibility.BUILTIN,
                                                  is_implicit=is_implicit)

    if is_function_identifier(id_node):
        # Functions can have the scope visibility only explicit global or
        # implicit global or explicit script local. So a function have implicit
        # scope visibility is always global function.
        return _create_identifier_visibility_hint(ScopeVisibility.GLOBAL_LIKE,
                                                  is_implicit=is_implicit)

    if not context_scope:
        # We cannot detect implicit scope visibility if context scope is not
        # specified.
        return _create_identifier_visibility_hint(ScopeVisibility.UNANALYZABLE)

    # Implicit scope variable will be resolved as a global variable when
    # current scope is script local. Otherwise be a function local variable.
    current_scope_visibility = context_scope['scope_visibility']
    if current_scope_visibility is ScopeVisibility.SCRIPT_LOCAL:
        return _create_identifier_visibility_hint(ScopeVisibility.GLOBAL_LIKE,
                                                  is_implicit=is_implicit)

    return _create_identifier_visibility_hint(ScopeVisibility.FUNCTION_LOCAL,
                                              is_implicit=is_implicit)


def _get_explicit_scope_visibility(id_node):
    # See :help internal-variables
    scope_prefix = id_node['value'][0:2]

    if is_function_identifier(id_node) and is_declarative_identifier(id_node):
        return FunctionDeclarationIdentifierScopePrefixToScopeVisibility.get(scope_prefix)
    else:
        return VariableIdentifierScopePrefixToScopeVisibility.get(scope_prefix)


def _create_identifier_visibility_hint(visibility, is_implicit=False):
    return {
        'scope_visibility': visibility,
        'is_implicit': is_implicit,
    }



AutoCmdEvents = {
    'BUFADD': True,
    'BUFCREATE': True,
    'BUFDELETE': True,
    'BUFENTER': True,
    'BUFFILEPOST': True,
    'BUFFILEPRE': True,
    'BUFHIDDEN': True,
    'BUFLEAVE': True,
    'BUFNEW': True,
    'BUFNEWFILE': True,
    'BUFREAD': True,
    'BUFREADCMD': True,
    'BUFREADPOST': True,
    'BUFREADPRE': True,
    'BUFUNLOAD': True,
    'BUFWINENTER': True,
    'BUFWINLEAVE': True,
    'BUFWIPEOUT': True,
    'BUFWRITE': True,
    'BUFWRITECMD': True,
    'BUFWRITEPOST': True,
    'BUFWRITEPRE': True,
    'CMDUNDEFINED': True,
    'CMDWINENTER': True,
    'CMDWINLEAVE': True,
    'COLORSCHEME': True,
    'COMPLETEDONE': True,
    'CURSORHOLD': True,
    'CURSORHOLDI': True,
    'CURSORMOVED': True,
    'CURSORMOVEDI': True,
    'ENCODINGCHANGED': True,
    'FILEAPPENDCMD': True,
    'FILEAPPENDPOST': True,
    'FILEAPPENDPRE': True,
    'FILECHANGEDRO': True,
    'FILECHANGEDSHELL': True,
    'FILECHANGEDSHELLPOST': True,
    'FILEREADCMD': True,
    'FILEREADPOST': True,
    'FILEREADPRE': True,
    'FILETYPE': True,
    'FILEWRITECMD': True,
    'FILEWRITEPOST': True,
    'FILEWRITEPRE': True,
    'FILTERREADPOST': True,
    'FILTERREADPRE': True,
    'FILTERWRITEPOST': True,
    'FILTERWRITEPRE': True,
    'FOCUSGAINED': True,
    'FOCUSLOST': True,
    'FUNCUNDEFINED': True,
    'GUIENTER': True,
    'GUIFAILED': True,
    'INSERTCHANGE': True,
    'INSERTCHARPRE': True,
    'INSERTENTER': True,
    'INSERTLEAVE': True,
    'MENUPOPUP': True,
    'QUICKFIXCMDPOST': True,
    'QUICKFIXCMDPRE': True,
    'QUITPRE': True,
    'REMOTEREPLY': True,
    'SESSIONLOADPOST': True,
    'SHELLCMDPOST': True,
    'SHELLFILTERPOST': True,
    'SOURCECMD': True,
    'SOURCEPRE': True,
    'SPELLFILEMISSING': True,
    'STDINREADPOST': True,
    'STDINREADPRE': True,
    'SWAPEXISTS': True,
    'SYNTAX': True,
    'TABENTER': True,
    'TABLEAVE': True,
    'TERMCHANGED': True,
    'TERMRESPONSE': True,
    'TEXTCHANGED': True,
    'TEXTCHANGEDI': True,
    'USER': True,
    'VIMENTER': True,
    'VIMLEAVE': True,
    'VIMLEAVEPRE': True,
    'VIMRESIZED': True,
    'WINENTER': True,
    'WINLEAVE': True,
}



from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy
from vint.ast.plugin.scope_plugin import ExplicityOfScopeVisibility


@register_policy
class ProhibitImplicitScopeVariable(AbstractPolicy):
    def __init__(self):
        super(ProhibitImplicitScopeVariable, self).__init__()
        self.reference = 'Anti-pattern of vimrc (Scope of identifier)'
        self.level = Level.STYLE_PROBLEM


    def listen_node_types(self):
        return [NodeType.IDENTIFIER]


    def is_valid(self, identifier, lint_context):
        """ Whether the identifier has a scope prefix. """

        scope_plugin = lint_context['plugins']['scope']
        explicity = scope_plugin.get_explicity_of_scope_visibility(identifier)

        is_valid = (scope_plugin.is_function_identifier(identifier) or
                    explicity is not ExplicityOfScopeVisibility.IMPLICIT)

        if not is_valid:
            self._make_description(identifier, scope_plugin)

        return is_valid


    def _make_description(self, identifier, scope_plugin):
        self.description = 'Make the scope explicit like `{good_example}`'.format(
            good_example=scope_plugin.normalize_variable_name(identifier)
        )



import re
from vint.linting.config.config_source import ConfigSource
from vint.ast.node_type import NodeType

CONFIG_COMMENT_PATTERN = re.compile(r'^\s*vint:\s*')
POLICY_SWITCH_PATTERN = re.compile(r'[-\+]\S+')


class ConfigCommentSource(ConfigSource):
    """ A class for ConfigCommentSourcees.
    This class provide to change config by modeline-like config comments as
    follow:

        " vint: -PolicyA
        " vint: +PolicyA
        " vint: -PolicyA +PolicyB

    Prefix vint: means that the comment is a config comment. And, +PolicyName
    means to enable the policy, and -PolicyName means to disable.
    """
    def __init__(self):
        self._config_dict = {
            'policies': {}
        }


    def get_config_dict(self):
        return self._config_dict


    def is_requesting_update(self, node):
        return self._is_config_comment(node)


    def update_by_node(self, node):
        comment_content = node['str']

        config_dict = self._config_dict
        config_dict['policies'] = self._parse_config_comment(comment_content)


    def _is_config_comment(self, node):
        if NodeType(node['type']) is not NodeType.COMMENT:
            return False

        comment_content = node['str']
        return CONFIG_COMMENT_PATTERN.match(comment_content) is not None


    def _parse_config_comment(self, comment_content):
        striped_comment_content = CONFIG_COMMENT_PATTERN.sub('', comment_content)

        policy_switches = POLICY_SWITCH_PATTERN.findall(striped_comment_content)

        config_dict = {}
        for policy_switch in policy_switches:
            policy_name = policy_switch[1:]
            is_enabling_switch = policy_switch[0] == '+'

            config_dict[policy_name] = {
                'enabled': is_enabling_switch
            }

        return config_dict



from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy


@register_policy
class ProhibitInvalidMapCall(AbstractPolicy):
    def __init__(self):
        super(ProhibitInvalidMapCall, self).__init__()
        self.description = 'Number of arguments for map() must be 2 (if not, it will throw E118 or E119)'
        self.reference = ':help map()'
        self.level = Level.ERROR


    def listen_node_types(self):
        return [NodeType.CALL]


    def is_valid(self, node, lint_context):
        left_node = node['left']

        if NodeType(left_node['type']) != NodeType.IDENTIFIER:
            return True

        if left_node['value'] != 'map':
            return True

        args = node['rlist']
        return len(args) == 2



import importlib
import pkgutil
from pathlib import Path

from vint.linting.cli import CLI
import logging


LOG_FORMAT = 'vint %(levelname)s: %(message)s'


def init_logger():
    logging.basicConfig(format=LOG_FORMAT)


def init_linter():
    import_all_policies()


def init_cli():
    cli = CLI()
    cli.start()


def import_all_policies():
    """ Import all policies that were registered by vint.linting.policy_registry.

    Dynamic policy importing is comprised of the 3 steps
      1. Try to import all policy modules (then we can't know what policies exist)
      2. In policy module, register itself by using vint.linting.policy_registry
      3. After all policies registered by itself, we can get policy classes
    """
    pkg_name = _get_policy_package_name_for_test()
    pkg_path_list = pkg_name.split('.')

    pkg_path = str(Path(_get_vint_root(), *pkg_path_list).resolve())

    for _, module_name, is_pkg in pkgutil.iter_modules([pkg_path]):
        if not is_pkg:
            module_fqn = pkg_name + '.' + module_name
            logging.debug('Loading the policy module: `{fqn}`'.format(fqn=module_fqn))
            importlib.import_module(module_fqn)


def _get_vint_root():
    return Path(__file__).parent.parent


def _get_policy_package_name_for_test():
    """ Test hook method that returns a package name for policy modules. """
    return 'vint.linting.policy'



from vint.ast.node_type import NodeType

SKIP_CHILDREN = 'SKIP_CHILDREN'


def for_each(func, nodes):
    """ Calls func for each the specified nodes. """
    for node in nodes:
        call_if_def(func, node)


def for_each_deeply(func, node_lists):
    """ Calls func for each the specified nodes. """
    for nodes in node_lists:
        for_each(func, nodes)


def call_if_def(func, node):
    """ Calls func if the node is defined.
    VimLParser return an empty array if a child node is not defined.
    """
    if hasattr(node, 'type'):
        func(node)


ChildNodeAccessor = {
    'NODE': call_if_def,
    'LIST': for_each,
    'NESTED_LIST': for_each_deeply,
}

ChildType = {
    'LEFT': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'left',
    },
    'RIGHT': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'right',
    },
    'COND': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'cond',
    },
    'REST': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'rest',
    },
    'LIST': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'list',
    },
    'RLIST': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'rlist',
    },
    'BODY': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'body',
    },
    'LIST_VALUES': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'value',
    },
    'DICT_ENTRIES': {
        'accessor': ChildNodeAccessor['NESTED_LIST'],
        'property_name': 'value',
    },
    'CURLYNAME_VALUES': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'value',
    },

    'ELSEIF': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'elseif',
    },
    'ELSE': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'else_',
    },
    'ENDIF': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'endif',
    },
    'ENDWHILE': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'endwhile',
    },
    'ENDFOR': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'endfor',
    },
    'CATCH': {
        'accessor': ChildNodeAccessor['LIST'],
        'property_name': 'catch',
    },
    'FINALLY': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'finally_',
    },
    'ENDTRY': {
        'accessor': ChildNodeAccessor['NODE'],
        'property_name': 'endtry',
    },
}

ChildNodeAccessorMap = {
    NodeType.TOPLEVEL: [ChildType['BODY']],
    NodeType.COMMENT: [],
    NodeType.EXCMD: [],
    NodeType.FUNCTION: [ChildType['LEFT'], ChildType['RLIST'], ChildType['BODY']],
    NodeType.ENDFUNCTION: [],
    NodeType.DELFUNCTION: [ChildType['LEFT']],
    NodeType.RETURN: [ChildType['LEFT']],
    NodeType.EXCALL: [ChildType['LEFT']],
    NodeType.LET: [ChildType['LEFT'], ChildType['LIST'], ChildType['REST'], ChildType['RIGHT']],
    NodeType.UNLET: [ChildType['LIST']],
    NodeType.LOCKVAR: [ChildType['LIST']],
    NodeType.UNLOCKVAR: [ChildType['LIST']],
    NodeType.IF: [ChildType['COND'], ChildType['BODY'], ChildType['ELSEIF'], ChildType['ELSE'], ChildType['ENDIF']],
    NodeType.ELSEIF: [ChildType['COND'], ChildType['BODY']],
    NodeType.ELSE: [ChildType['BODY']],
    NodeType.ENDIF: [],
    NodeType.WHILE: [ChildType['COND'], ChildType['BODY'], ChildType['ENDWHILE']],
    NodeType.ENDWHILE: [],
    NodeType.FOR: [ChildType['BODY'], ChildType['LEFT'], ChildType['LIST'], ChildType['REST'], ChildType['RIGHT'], ChildType['ENDFOR']],
    NodeType.ENDFOR: [],
    NodeType.CONTINUE: [],
    NodeType.BREAK: [],
    NodeType.TRY: [ChildType['BODY'], ChildType['CATCH'], ChildType['FINALLY'], ChildType['ENDTRY']],
    NodeType.CATCH: [ChildType['BODY']],
    NodeType.FINALLY: [ChildType['BODY']],
    NodeType.ENDTRY: [],
    NodeType.THROW: [ChildType['LEFT']],
    NodeType.ECHO: [ChildType['LIST']],
    NodeType.ECHON: [ChildType['LIST']],
    NodeType.ECHOHL: [],
    NodeType.ECHOMSG: [ChildType['LIST']],
    NodeType.ECHOERR: [ChildType['LIST']],
    NodeType.EXECUTE: [ChildType['LIST']],
    NodeType.TERNARY: [ChildType['COND'], ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.OR: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.AND: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.EQUAL: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.EQUALCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.EQUALCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NEQUAL: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NEQUALCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NEQUALCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.GREATER: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.GREATERCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.GREATERCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.GEQUAL: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.GEQUALCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.GEQUALCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SMALLER: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SMALLERCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SMALLERCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SEQUAL: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SEQUALCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SEQUALCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.MATCH: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.MATCHCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.MATCHCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NOMATCH: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NOMATCHCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NOMATCHCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.IS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.ISCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.ISCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.ISNOT: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.ISNOTCI: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.ISNOTCS: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.ADD: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SUBTRACT: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.CONCAT: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.MULTIPLY: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.DIVIDE: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.REMAINDER: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NOT: [ChildType['LEFT']],
    NodeType.MINUS: [ChildType['LEFT']],
    NodeType.PLUS: [ChildType['LEFT']],
    NodeType.SUBSCRIPT: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.SLICE: [ChildType['LEFT'], ChildType['RLIST']],
    NodeType.CALL: [ChildType['LEFT'], ChildType['RLIST']],
    NodeType.DOT: [ChildType['LEFT'], ChildType['RIGHT']],
    NodeType.NUMBER: [],
    NodeType.STRING: [],
    NodeType.LIST: [ChildType['LIST_VALUES']],
    NodeType.DICT: [ChildType['DICT_ENTRIES']],
    NodeType.NESTING: [ChildType['LEFT']],
    NodeType.OPTION: [],
    NodeType.IDENTIFIER: [],
    NodeType.CURLYNAME: [ChildType['CURLYNAME_VALUES']],
    NodeType.ENV: [],
    NodeType.REG: [],
    NodeType.CURLYNAMEPART: [],
    NodeType.CURLYNAMEEXPR: [],
    NodeType.LAMBDA: [],
}


class UnknownNodeTypeException(BaseException):
    def __init__(self, node_type):
        self.node_type = node_type

    def __str__(self):
        node_type_name = self.node_type
        return 'Unknown node type: `{node_type}`'.format(node_type=node_type_name)


_traverser_extensions = []


def register_traverser_extension(handler):
    """ Registers the specified function to traverse into extended child nodes.
    """
    _traverser_extensions.append(handler)


def traverse(node, on_enter=None, on_leave=None):
    """ Traverses the specified Vim script AST node (depth first order).
    The on_enter/on_leave handler will be called with the specified node and
    the children. You can skip traversing child nodes by returning
    SKIP_CHILDREN.
    """
    node_type = NodeType(node['type'])

    if node_type not in ChildNodeAccessorMap:
        raise UnknownNodeTypeException(node_type)

    if on_enter:
        should_traverse_children = on_enter(node) is not SKIP_CHILDREN
    else:
        should_traverse_children = True

    if should_traverse_children:
        for property_accessor in ChildNodeAccessorMap[node_type]:
            accessor_func = property_accessor['accessor']
            prop_name = property_accessor['property_name']

            accessor_func(lambda child_node: traverse(child_node, on_enter, on_leave),
                          node[prop_name])

        for handler in _traverser_extensions:
            handler(node, on_enter=on_enter, on_leave=on_leave)

    if on_leave:
        on_leave(node)



from pathlib import Path
from vint.asset import get_asset_path
from vint.linting.config.config_file_source import ConfigFileSource

PROJECT_CONFIG_FILENAMES = [
    '.vintrc.yaml',
    '.vintrc.yml',
    '.vintrc',
]


class ConfigProjectSource(ConfigFileSource):
    def get_file_path(self, env):
        cwd = Path(env['cwd'])
        path_list_to_search = [cwd] + list(cwd.parents)

        for project_path in path_list_to_search:
            for basename in PROJECT_CONFIG_FILENAMES:
                proj_conf_path_tmp = project_path / basename

                if proj_conf_path_tmp.is_file():
                    return proj_conf_path_tmp

        return get_asset_path('void_config.yaml')



from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy.reference.googlevimscriptstyleguide import get_reference_source
from vint.linting.policy_registry import register_policy


@register_policy
class ProhibitNoAbortFunction(AbstractPolicy):
    def __init__(self):
        super(ProhibitNoAbortFunction, self).__init__()
        self.description = 'Use the abort attribute for functions in autoload'
        self.reference = get_reference_source('FUNCTIONS')
        self.level = Level.WARNING


    def listen_node_types(self):
        return [NodeType.FUNCTION]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        This policy prohibits functions in autoload that have no 'abort' or bang
        """

        if 'autoload' not in lint_context['path'].parts:
            return True

        has_bang = node['ea']['forceit'] != 0
        has_abort = node['attr']['abort'] != 0

        return has_bang and has_abort



from vint.linting.config.config_source import ConfigSource


class ConfigDictSource(ConfigSource):
    def __init__(self, config_dict):
        self._config_dict = config_dict


    def get_config_dict(self):
        return self._config_dict



import re
import logging
from pathlib import Path
from vint._bundles import vimlparser
from vint.encodings.decoder import EncodingDetectionError
from vint.ast.parsing import Parser
from vint.ast.node_type import NodeType
from vint.ast.traversing import traverse
from vint.ast.plugin.scope_plugin import ScopePlugin
from vint.linting.config.config_container import ConfigContainer
from vint.linting.config.config_dict_source import ConfigDictSource
from vint.linting.config.config_comment_source import ConfigCommentSource
from vint.linting.config.config_util import get_config_value
from vint.linting.level import Level


class Linter(object):
    """ A class for Linters.
    This class provides a linting algorithm that know how to lint by Policy
    class, and how to manage enabled policies by using Config classes.

    The class manage what policies are enabled by using the Config class.
    The config should be refreshed when a config comment is found.
    Then the Linter class notify refresh request to the Config class in
    traversing.

    The Linter class lint by event-driven traversing.
    Enabled policies request to call their when the nodes necesary to validate
    are found in traversing. The Linter class collect violations that are
    returned by policies.
    """
    def __init__(self, policy_set, config_dict_global):
        self._plugins = {
            'scope': ScopePlugin(),
        }
        self._policy_set = policy_set

        self._config_comment_source = ConfigCommentSource()
        self._config = self._decorate_config(config_dict_global,
                                             self._config_comment_source)

        self._parser = self.build_parser()

        self._listeners_map = {}


    def build_parser(self):
        config_dict = self._config.get_config_dict()
        enable_neovim = get_config_value(config_dict, ['cmdargs', 'env', 'neovim'], False)

        parser = Parser(self._plugins, enable_neovim=enable_neovim)
        return parser


    def _decorate_config(self, config_dict_global, config_comment_source):
        config_dict_source = ConfigDictSource(config_dict_global)

        config_container = ConfigContainer(config_dict_source,
                                           config_comment_source)

        return config_container


    def _parse_vimlparser_error(self, err_message):
        match = re.match(r'vimlparser: (?P<description>.*): line (?P<line_number>\d+) col (?P<column_number>\d+)$', err_message)
        return match.groupdict()


    def _create_parse_error(self, path, err_message):
        parser_error = self._parse_vimlparser_error(err_message)
        return {
            'name': 'SyntaxError',
            'level': Level.ERROR,
            'description': parser_error['description'],
            'reference': 'vim-jp/vim-vimlparser',
            'position': {
                'line': int(parser_error['line_number']),
                'column': int(parser_error['column_number']),
                'path': Path(path),
            },
        }


    def _create_decoding_error(self, path, err_message):
        return {
            'name': 'DecodingError',
            'level': Level.ERROR,
            'description': err_message,
            'reference': 'no reference',
            'position': {
                'line': 0,
                'column': 0,
                'path': Path(path),
            },
        }


    def _log_file_path_to_lint(self, file_path):
        msg = 'checking: `{file_path}`'.format(file_path=file_path)
        logging.debug(msg)


    def lint_file(self, path):
        self._log_file_path_to_lint(path)

        try:
            root_ast = self._parser.parse_file(path)
        except vimlparser.VimLParserException as exception:
            parse_error = self._create_parse_error(path, str(exception))
            return [parse_error]
        except EncodingDetectionError as exception:
            decoding_error = self._create_decoding_error(path, str(exception))
            return [decoding_error]

        self._traverse(root_ast, path)

        return self._violations


    def lint_text(self, text):
        self._log_file_path_to_lint('stdin')

        try:
            root_ast = self._parser.parse(text)
        except vimlparser.VimLParserException as exception:
            parse_error = self._create_parse_error('stdin', str(exception))
            return [parse_error]
        except EncodingDetectionError as exception:
            decoding_error = self._create_decoding_error('stdin', str(exception))
            return [decoding_error]

        self._traverse(root_ast, 'stdin')

        return self._violations


    def _traverse(self, root_ast, path):
        self._prepare_for_traversal()

        lint_context = {
            'path': path,
            'root_node': root_ast,
            'stack_trace': [],
            'plugins': self._plugins,
            'config': self._config.get_config_dict(),
        }

        traverse(root_ast,
                 on_enter=lambda node: self._handle_enter(node, lint_context),
                 on_leave=lambda node: self._handle_leave(node, lint_context))


    def _prepare_for_traversal(self):
        self._violations = []
        self._update_listeners_map()


    def _handle_enter(self, node, lint_context):
        self._refresh_policies_if_necessary(node)
        self._fire_listeners(node, lint_context)

        lint_context['stack_trace'].append(node)


    def _handle_leave(self, node, lint_context):
        lint_context['stack_trace'].pop()


    def _fire_listeners(self, node, lint_context):
        node_type = NodeType(node['type'])

        if node_type not in self._listeners_map:
            return

        listening_policies = self._listeners_map[node_type]

        for listening_policy in listening_policies:
            violation = listening_policy.get_violation_if_found(node, lint_context)

            if violation is not None:
                self._violations.append(violation)


    def _refresh_policies_if_necessary(self, node):
        config_comment_source = self._config_comment_source

        if not config_comment_source.is_requesting_update(node):
            return

        self._refresh_policies(node)


    def _refresh_policies(self, node):
        config_comment_source = self._config_comment_source
        config_comment_source.update_by_node(node)

        self._update_listeners_map()


    def _update_enabled_policies(self):
        policy_set = self._policy_set
        config = self._config

        config_dict = config.get_config_dict()
        policy_set.update_by_config(config_dict)


    def _update_listeners_map(self):
        self._update_enabled_policies()

        self._listeners_map = {}
        policy_set = self._policy_set

        for policy in policy_set.get_enabled_policies():
            listened_node_types = policy.listen_node_types()

            for listened_node_type in listened_node_types:
                if listened_node_type not in self._listeners_map:
                    self._listeners_map[listened_node_type] = [policy]
                else:
                    self._listeners_map[listened_node_type].append(policy)



import chardet
import sys

from vint.ast.node_type import NodeType
from vint.ast.traversing import traverse, SKIP_CHILDREN
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy


@register_policy
class ProhibitMissingScriptEncoding(AbstractPolicy):
    def __init__(self):
        super(ProhibitMissingScriptEncoding, self).__init__()
        self.description = 'Use scriptencoding when multibyte char exists'
        self.reference = ':help :scriptencoding'
        self.level = Level.WARNING

        self.has_scriptencoding = False


    def listen_node_types(self):
        return [NodeType.TOPLEVEL]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        This policy prohibit scriptencoding missing when multibyte char exists.
        """
        traverse(node, on_enter=self._check_scriptencoding)

        if self.has_scriptencoding:
            return True

        return not self._check_script_has_multibyte_char(lint_context)


    def _check_scriptencoding(self, node):
        # TODO: Use BREAK when implemented
        if self.has_scriptencoding:
            return SKIP_CHILDREN

        node_type = NodeType(node['type'])

        if node_type is not NodeType.EXCMD:
            return

        self.has_scriptencoding = node['str'].startswith('scripte')


    def _check_script_has_multibyte_char(self, lint_context):
        if self._is_from_stdin(lint_context):
            return self._check_stdin_has_multibyte_char()
        else:
            return self._check_file_has_multibyte_char(lint_context)


    def _is_from_stdin(self, lint_context):
        return lint_context['path'] == 'stdin'


    def _check_stdin_has_multibyte_char(self):
        sys.stdin.seek(0)

        try:
            sys.stdin.read().encode('ascii')
            return False
        except UnicodeError:
            return True

    def _check_file_has_multibyte_char(self, lint_context):
        # TODO: Use cache to make performance efficiency
        with lint_context['path'].open(mode='br') as f:
            byte_seq = f.read()
        return len(byte_seq) and chardet.detect(byte_seq)['encoding'] != 'ascii'



import yaml
import logging
from vint.linting.config.config_source import ConfigSource
from vint.linting.level import Level



class ConfigFileSource(ConfigSource):
    def __init__(self, env):
        config_file_path = self.get_file_path(env)

        with config_file_path.open() as file_obj:
            config_yaml = yaml.safe_load(file_obj)
            self._config_dict = self.convert_config_dict(config_yaml)


    def convert_config_dict(self, yaml_dict):
        if not yaml_dict:
            return {}

        if 'cmdargs' in yaml_dict:
            # Care empty hash, because it becomes None.
            if yaml_dict['cmdargs'] is None:
                yaml_dict['cmdargs'] = {}

            cmdargs_dict = yaml_dict['cmdargs']

            if 'severity' in cmdargs_dict:
                severity_name = cmdargs_dict['severity'].upper()
                severity = getattr(Level, severity_name, None)

                if not severity:
                    self._warn_invalid_severity(severity_name)
                    return yaml_dict

                cmdargs_dict['severity'] = severity

        return yaml_dict


    def _warn_invalid_severity(self, severity_name):
        possible_severities = ', '.join([possible_severity.name.lower()
                                         for possible_severity
                                         in list(Level)])

        logging.warning(('Severity `{severity_name}` is invalid. ' +
                         'Possible: {possible_severities}').format(
                             severity_name=severity_name,
                             possible_severities=possible_severities))


    def get_file_path(self, env):
        raise NotImplementedError()


    def get_config_dict(self):
        return self._config_dict



import sys
from argparse import ArgumentParser
from pathlib import PosixPath
import pkg_resources
import logging

from vint.linting.linter import Linter
from vint.linting.env import build_environment
from vint.linting.config.config_container import ConfigContainer
from vint.linting.config.config_cmdargs_source import ConfigCmdargsSource
from vint.linting.config.config_default_source import ConfigDefaultSource
from vint.linting.config.config_global_source import ConfigGlobalSource
from vint.linting.config.config_project_source import ConfigProjectSource
from vint.linting.policy_set import PolicySet
from vint.linting.formatter.formatter import Formatter
from vint.linting.formatter.json_formatter import JSONFormatter
from vint.linting.formatter.statistic_formatter import StatisticFormatter

class CLI(object):
    def start(self):
        env = self._build_env(sys.argv)
        self._validate(env)

        self._adjust_log_level(env)

        config_dict = self._build_config_dict(env)
        violations = self._lint_all(env, config_dict)

        parser = self._build_argparser()

        if len(violations) == 0:
            parser.exit(status=0)

        self._print_violations(violations, config_dict)
        parser.exit(status=1)


    def _validate(self, env):
        parser = self._build_argparser()
        paths_to_lint = env['file_paths']

        if len(paths_to_lint) == 0:
            logging.error('nothing to check')
            parser.print_help()
            parser.exit(status=1)

        if not self._should_read_from_stdin(env):
            for path_to_lint in paths_to_lint:
                if not path_to_lint.exists() or not path_to_lint.is_file():
                    logging.error('no such file or directory: `{path}`'.format(
                        path=str(path_to_lint)))
                    parser.exit(status=1)


    def _build_config_dict(self, env):
        config = ConfigContainer(
            ConfigDefaultSource(env),
            ConfigGlobalSource(env),
            ConfigProjectSource(env),
            ConfigCmdargsSource(env),
        )

        return config.get_config_dict()


    def _build_argparser(self):
        parser = ArgumentParser(prog='vint', description='Lint Vim script')

        parser.add_argument('-v', '--version', action='version', version=self._get_version())
        parser.add_argument('-V', '--verbose', action='store_const', const=True, help='output verbose message')
        parser.add_argument('-e', '--error', action='store_const', const=True, help='report only errors')
        parser.add_argument('-w', '--warning', action='store_const', const=True, help='report errors and warnings')
        parser.add_argument('-s', '--style-problem', action='store_const', const=True, help='report errors, warnings and style problems')
        parser.add_argument('-m', '--max-violations', type=int, help='limit max violations count')
        parser.add_argument('-c', '--color', action='store_const', const=True, help='colorize output when possible')
        parser.add_argument('--no-color', action='store_const', const=True, help='do not colorize output')
        parser.add_argument('-j', '--json', action='store_const', const=True, help='output json style')
        parser.add_argument('-t', '--stat', action='store_const', const=True, help='output statistic info')
        parser.add_argument('--enable-neovim', action='store_const', const=True, help='Enable Neovim syntax')
        parser.add_argument('-f', '--format', help='set output format')
        parser.add_argument('files', nargs='*', help='file or directory path to lint')

        return parser


    def _build_cmdargs(self, argv):
        """ Build command line arguments dict to use;
        - displaying usages
        - vint.linting.env.build_environment

        This method take an argv parameter to make function pure.
        """
        parser = self._build_argparser()
        namespace = parser.parse_args(argv[1:])

        cmdargs = vars(namespace)
        return cmdargs


    def _build_env(self, argv):
        """ Build an environment object.
        This method take an argv parameter to make function pure.
        """
        cmdargs = self._build_cmdargs(argv)
        env = build_environment(cmdargs)
        return env


    def _build_linter(self, config_dict):
        policy_set = PolicySet()
        linter = Linter(policy_set, config_dict)
        return linter


    def _lint_all(self, env, config_dict):
        paths_to_lint = env['file_paths']
        violations = []
        linter = self._build_linter(config_dict)

        if self._should_read_from_stdin(env):
            violations += linter.lint_text(sys.stdin.read())
        else:
            for file_path in paths_to_lint:
                violations += linter.lint_file(file_path)

        return violations

    def _should_read_from_stdin(self, env):
        return len(env['file_paths']) == 1 and PosixPath('-') in env['file_paths']


    def _get_formatter(self, config_dict):
        if 'cmdargs' not in config_dict:
            return Formatter(config_dict)

        cmdargs = config_dict['cmdargs']
        if 'json' in cmdargs and cmdargs['json']:
            return JSONFormatter(config_dict)
        elif 'stat' in cmdargs and cmdargs['stat']:
            return StatisticFormatter(config_dict)
        else:
            return Formatter(config_dict)


    def _print_violations(self, violations, config_dict):
        formatter = self._get_formatter(config_dict)
        output = formatter.format_violations(violations)

        print(output)


    def _get_version(self):
        # In unit tests, pkg_resources cannot find vim-vint.
        # So, I decided to return dummy version
        try:
            version = pkg_resources.require('vim-vint')[0].version
        except pkg_resources.DistributionNotFound:
            version = 'test_mode'

        return version


    def _adjust_log_level(self, env):
        cmdargs = env['cmdargs']

        is_verbose = cmdargs.get('verbose', False)
        log_level = logging.DEBUG if is_verbose else logging.WARNING

        logger = logging.getLogger()
        logger.setLevel(log_level)



from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy
from vint.ast.plugin.scope_plugin import ScopeVisibility


@register_policy
class ProhibitUnusedVariable(AbstractPolicy):
    def __init__(self):
        super(ProhibitUnusedVariable, self).__init__()
        self.reference = ':help E738'
        self.level = Level.WARNING


    def listen_node_types(self):
        return [NodeType.IDENTIFIER]


    def is_valid(self, identifier, lint_context):
        """ Whether the variables are used.
        This policy cannot determine the following node types:
          - Global identifier like nodes
            - ENV
            - REG
            - OPTION
          - Dynamic variables
            - CURLYNAME
            - SLICE
            - DOT
            - SUBSCRIPT
        """

        scope_plugin = lint_context['plugins']['scope']
        if not scope_plugin.is_unused_declarative_identifier(identifier):
            return True

        # Ignore global like variables.
        scope_visibility = scope_plugin.get_objective_scope_visibility(identifier)
        if (scope_visibility is ScopeVisibility.GLOBAL_LIKE or
                scope_visibility is ScopeVisibility.BUILTIN or
                scope_visibility is ScopeVisibility.UNANALYZABLE):
            return True

        self.description = 'Unused variable: {var_name}'.format(
            var_name=identifier['value'])
        return False



BuiltinIdentifierMap = {
    'abs': True,
    'acos': True,
    'add': True,
    'append': True,
    'argc': True,
    'argidx': True,
    'argv': True,
    'asin': True,
    'atan': True,
    'atan2': True,
    'browse': True,
    'browsedir': True,
    'bufexists': True,
    'buflisted': True,
    'bufloaded': True,
    'bufname': True,
    'bufnr': True,
    'bufwinnr': True,
    'byte2line': True,
    'byteidx': True,
    'call': True,
    'ceil': True,
    'changenr': True,
    'char2nr': True,
    'cindent': True,
    'clearmatches': True,
    'col': True,
    'complete': True,
    'complete_add': True,
    'complete_check': True,
    'confirm': True,
    'copy': True,
    'cos': True,
    'cosh': True,
    'count': True,
    'cscope_connection': True,
    'cursor': True,
    'deepcopy': True,
    'delete': True,
    'did_filetype': True,
    'diff_filler': True,
    'diff_hlID': True,
    'empty': True,
    'escape': True,
    'eval': True,
    'eventhandler': True,
    'executable': True,
    'exists': True,
    'exp': True,
    'expand': True,
    'extend': True,
    'feedkeys': True,
    'filereadable': True,
    'filewritable': True,
    'filter': True,
    'finddir': True,
    'findfile': True,
    'float2nr': True,
    'floor': True,
    'fmod': True,
    'fnameescape': True,
    'fnamemodify': True,
    'foldclosed': True,
    'foldclosedend': True,
    'foldlevel': True,
    'foldtext': True,
    'foldtextresult': True,
    'foreground': True,
    'function': True,
    'garbagecollect': True,
    'get': True,
    'getbufline': True,
    'getbufvar': True,
    'getchar': True,
    'getcharmod': True,
    'getcmdline': True,
    'getcmdpos': True,
    'getcmdtype': True,
    'getcwd': True,
    'getfontname': True,
    'getfperm': True,
    'getfsize': True,
    'getftime': True,
    'getftype': True,
    'getline': True,
    'getloclist': True,
    'getmatches': True,
    'getpid': True,
    'getpos': True,
    'getqflist': True,
    'getreg': True,
    'getregtype': True,
    'gettabvar': True,
    'gettabwinvar': True,
    'getwinposx': True,
    'getwinposy': True,
    'getwinvar': True,
    'glob': True,
    'globpath': True,
    'has': True,
    'has_key': True,
    'haslocaldir': True,
    'hasmapto': True,
    'histadd': True,
    'histdel': True,
    'histget': True,
    'histnr': True,
    'hlID': True,
    'hlexists': True,
    'hostname': True,
    'iconv': True,
    'indent': True,
    'index': True,
    'input': True,
    'inputdialog': True,
    'inputlist': True,
    'inputrestore': True,
    'inputsave': True,
    'inputsecret': True,
    'insert': True,
    'isdirectory': True,
    'islocked': True,
    'items': True,
    'join': True,
    'keys': True,
    'len': True,
    'libcall': True,
    'libcallnr': True,
    'line': True,
    'line2byte': True,
    'lispindent': True,
    'localtime': True,
    'log': True,
    'log10': True,
    'map': True,
    'maparg': True,
    'mapcheck': True,
    'match': True,
    'matchadd': True,
    'matcharg': True,
    'matchdelete': True,
    'matchend': True,
    'matchlist': True,
    'matchstr': True,
    'max': True,
    'min': True,
    'mkdir': True,
    'mode': True,
    'nextnonblank': True,
    'nr2char': True,
    'pathshorten': True,
    'pow': True,
    'prevnonblank': True,
    'printf': True,
    'pumvisible': True,
    'range': True,
    'readfile': True,
    'reltime': True,
    'reltimestr': True,
    'remote_expr': True,
    'remote_foreground': True,
    'remote_peek': True,
    'remote_read': True,
    'remote_send': True,
    'remove': True,
    'rename': True,
    'repeat': True,
    'resolve': True,
    'reverse': True,
    'round': True,
    'search': True,
    'searchdecl': True,
    'searchpair': True,
    'searchpairpos': True,
    'searchpos': True,
    'server2client': True,
    'serverlist': True,
    'setbufvar': True,
    'setcmdpos': True,
    'setline': True,
    'setloclist': True,
    'setmatches': True,
    'setpos': True,
    'setqflist': True,
    'setreg': True,
    'settabvar': True,
    'settabwinvar': True,
    'setwinvar': True,
    'shellescape': True,
    'simplify': True,
    'sin': True,
    'sinh': True,
    'sort': True,
    'soundfold': True,
    'spellbadword': True,
    'spellsuggest': True,
    'split': True,
    'sqrt': True,
    'str2float': True,
    'str2nr': True,
    'strchars': True,
    'strdisplaywidth': True,
    'strftime': True,
    'stridx': True,
    'string': True,
    'strlen': True,
    'strpart': True,
    'strridx': True,
    'strtrans': True,
    'strwidth': True,
    'submatch': True,
    'substitute': True,
    'synID': True,
    'synIDattr': True,
    'synIDtrans': True,
    'synconcealed': True,
    'synstack': True,
    'system': True,
    'tabpagebuflist': True,
    'tabpagenr': True,
    'tabpagewinnr': True,
    'tagfiles': True,
    'taglist': True,
    'tan': True,
    'tanh': True,
    'tempname': True,
    'tolower': True,
    'toupper': True,
    'tr': True,
    'trunc': True,
    'type': True,
    'undofile': True,
    'undotree': True,
    'values': True,
    'virtcol': True,
    'visualmode': True,
    'winbufnr': True,
    'wincol': True,
    'winheight': True,
    'winline': True,
    'winnr': True,
    'winrestcmd': True,
    'winrestview': True,
    'winsaveview': True,
    'winwidth': True,
    'writefile': True,
}



from typing import TypeVar, List, Callable
from functools import reduce
from operator import add

T = TypeVar('T')
S = TypeVar('S')


def flatten(l):
    # type: (List[List[T]]) -> List[T]
    return reduce(add, l, [])


def flat_map(f, l):
    # type: (Callable[[S], List[T]], List[S]) -> List[T]
    return flatten([f(x) for x in l])


import re

VIM_SCRIPT_FILE_NAME_PATTERNS = r'(?:^[\._]g?vimrc$|.*\.vim$)'


def find_vim_script(file_paths):
    vim_script_file_paths = []

    for file_path in file_paths:
        if file_path.is_dir():
            vim_script_file_paths += _find_vim_script_into_dir(file_path)
            continue

        vim_script_file_paths.append(file_path)

    return vim_script_file_paths


def _find_vim_script_into_dir(dir_path):
    if not dir_path.is_dir():
        return []

    vim_script_file_paths = []

    for file_path in dir_path.iterdir():
        if _is_vim_script(file_path):
            vim_script_file_paths.append(file_path)
            continue

        if file_path.is_dir():
            vim_script_file_paths += _find_vim_script_into_dir(file_path)
            continue

    return vim_script_file_paths


def _is_vim_script(file_path):
    if not file_path.is_file():
        return False

    file_name = file_path.name
    return bool(re.match(VIM_SCRIPT_FILE_NAME_PATTERNS, file_name))



from vint.ast.plugin.scope_plugin.scope_detector import (
    detect_scope_visibility,
    normalize_variable_name,
    is_builtin_variable,
)
from vint.ast.plugin.scope_plugin.scope_linker import ScopeLinker
from vint.ast.plugin.scope_plugin.identifier_classifier import (
    IdentifierClassifier,
    is_function_identifier,
)


REACHABILITY_FLAG = 'VINT:is_reachable'
REFERECED_FLAG = 'VINT:is_referenced'


class ReferenceReachabilityTester(object):
    """ A class for tester for reference reachabilities. """

    class TwoWayScopeReferenceAttacher(object):
        """ A class for AST processor to do attach to way reference between
        a parent scope and the child scopes.
        """

        @classmethod
        def attach(cls, root_scope_tree):
            root_scope_tree['parent_scope'] = None

            return cls._attach_recursively(root_scope_tree)


        @classmethod
        def _attach_recursively(cls, scope_tree):
            for child_scope in scope_tree['child_scopes']:
                child_scope['parent_scope'] = scope_tree
                cls._attach_recursively(child_scope)

            return scope_tree


    def process(self, ast):
        scope_linker = ScopeLinker()
        scope_linker.process(ast)

        id_collector = IdentifierClassifier.IdentifierCollector()
        classified_id_group = id_collector.collect_identifiers(ast)
        dec_id_nodes = classified_id_group['static_declaring_identifiers']
        ref_id_nodes = classified_id_group['static_referencing_identifiers']

        self._scope_tree = scope_linker.scope_tree
        self._link_registry = scope_linker.link_registry

        # Attach a parent_scope accessor to the scope tree
        ReferenceReachabilityTester.TwoWayScopeReferenceAttacher.attach(self._scope_tree)

        # Reset REFERECED_FLAG to False
        for dec_id_node in dec_id_nodes:
            dec_id_node[REFERECED_FLAG] = False

        for ref_id_node in ref_id_nodes:
            is_reachable = self.check_reachability(ref_id_node)
            ref_id_node[REACHABILITY_FLAG] = is_reachable


    def get_objective_scope_visibility(self, node):
        """ Returns a objective scope visibility. """
        context_scope = self._link_registry.get_scope_by_referencing_identifier(node)
        return detect_scope_visibility(node, context_scope)['scope_visibility']


    def _reset_referenced_flag(self, scope_tree):
        for child_scope in scope_tree['child_scopes']:
            for functions in child_scope['functions'].values():
                for function in functions:
                    function[REFERECED_FLAG] = False

            for variables in child_scope['variables'].values():
                for variable in variables:
                    variable[REFERECED_FLAG] = False

            self._reset_referenced_flag(child_scope)


    def check_reachability(self, ref_id_node):
        scope = self._link_registry.get_context_scope_by_identifier(ref_id_node)
        var_name = normalize_variable_name(ref_id_node, scope)
        is_func_id = is_function_identifier(ref_id_node)

        while scope is not None:
            if is_func_id:
                functions_list = scope['functions']
                if var_name in functions_list:
                    # The function is found in the symbol table for functions.
                    for variable in functions_list[var_name]:
                        declaring_id_node = self._link_registry\
                            .get_declarative_identifier_by_variable(variable)
                        declaring_id_node[REFERECED_FLAG] = True

                    return True
                else:
                    # We can access the function via a variable function
                    # reference if the function not found from the symbol table
                    # for functions. So we should check the symbol table for
                    # variables to search the function reference.
                    pass

            variables_list = scope['variables']
            if var_name in variables_list:
                # The variable or function reference found in the symbol table
                # for variables.
                for variable in variables_list[var_name]:
                    declaring_id_node = self._link_registry\
                        .get_declarative_identifier_by_variable(variable)
                    declaring_id_node[REFERECED_FLAG] = True

                return True

            scope = scope['parent_scope']

        # If it is builtin, it will be used by Vim.
        return is_builtin_variable(ref_id_node)


def is_reference_identifier(node):
    return REACHABILITY_FLAG in node


def is_reachable_reference_identifier(node):
    return node.get(REACHABILITY_FLAG, False)


def is_declarative_identifier(node):
    return REFERECED_FLAG in node


def is_referenced_declarative_identifier(node):
    return node.get(REFERECED_FLAG, False)



from vint.ast.plugin.scope_plugin.reference_reachability_tester import (
    ReferenceReachabilityTester,
    is_reference_identifier as _is_reference_identifier,
    is_declarative_identifier as _is_declarative_identifier,
    is_reachable_reference_identifier as _is_reachable_reference_identifier,
    is_referenced_declarative_identifier as _is_referenced_declarative_identifier,
)
from vint.ast.plugin.scope_plugin.scope_detector import (
    ScopeVisibility as _ScopeVisibility,
    ExplicityOfScopeVisibility as _ExplicityOfScopeVisibility,
    detect_scope_visibility as _detect_scope_visibility,
    get_explicity_of_scope_visibility as _get_explicity_of_scope_visibility,
    normalize_variable_name as _normalize_variable_name,
)
from vint.ast.plugin.scope_plugin.identifier_classifier import (
    is_autoload_identifier as _is_autoload_identifier,
    is_function_identifier as _is_function_identifier,
)


# Expose to out of ScopePlugin
ScopeVisibility = _ScopeVisibility
ExplicityOfScopeVisibility = _ExplicityOfScopeVisibility


class ScopePlugin(object):
    def __init__(self):
        self._ref_tester = ReferenceReachabilityTester()


    def process(self, ast):
        processed_ast = self._ref_tester.process(ast)
        return processed_ast


    def _get_link_registry(self):
        # NOTE: This is a hack for performance. We should build LinkRegistry
        # by this method if ReferenceReachabilityTester hide the link_registry.
        return self._ref_tester._link_registry


    def is_unreachable_reference_identifier(self, node):
        return _is_reference_identifier(node) \
            and not _is_reachable_reference_identifier(node)


    def is_unused_declarative_identifier(self, node):
        return _is_declarative_identifier(node) \
            and not _is_referenced_declarative_identifier(node)


    def is_autoload_identifier(self, node):
        return _is_autoload_identifier(node)


    def is_function_identifier(self, node):
        return _is_function_identifier(node)

    def get_objective_scope_visibility(self, node):
        link_registry = self._get_link_registry()
        context_scope = link_registry.get_context_scope_by_identifier(node)
        scope_visibility_hint = _detect_scope_visibility(node, context_scope)
        return scope_visibility_hint['scope_visibility']


    def get_explicity_of_scope_visibility(self, node):
        return _get_explicity_of_scope_visibility(node)


    def normalize_variable_name(self, node):
        link_registry = self._get_link_registry()
        context_scope = link_registry.get_context_scope_by_identifier(node)
        return _normalize_variable_name(node, context_scope)



import re
from vint.ast.node_type import NodeType
from vint.linting.level import Level
from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy_registry import register_policy
from vint.ast.dictionary.abbreviations import (
    Abbreviations,
    AbbreviationsIncludingInvertPrefix,
)


SetCommandFamily = {
    'set': True,
    'setlocal': True,
    'setglobal': True,
}


@register_policy
class ProhibitAbbreviationOption(AbstractPolicy):
    def __init__(self):
        super(ProhibitAbbreviationOption, self).__init__()
        self.description = 'Use the full option name instead of the abbreviation'
        self.reference = ':help option-summary'
        self.level = Level.STYLE_PROBLEM

        self.was_scriptencoding_found = False
        self.has_encoding_opt_after_scriptencoding = False


    def listen_node_types(self):
        return [NodeType.EXCMD, NodeType.OPTION]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid.

        Abbreviation options are invalid.
        """

        node_type = NodeType(node['type'])

        if node_type is NodeType.OPTION:
            # Remove & at head
            option_name = node['value'][1:]
            is_valid = option_name not in Abbreviations

            if not is_valid:
                self._make_description_by_option_name(option_name)

            return is_valid

        excmd_node = node
        is_set_cmd = excmd_node['ea']['cmd'].get('name') in SetCommandFamily

        if not is_set_cmd:
            return True

        option_expr = excmd_node['str'].split()[1]
        # Care `:set ft=vim` and `:set cpo&vim`, ...
        option_name = re.match(r'[a-z]+', option_expr).group(0)

        # After a "set" command, we can add an invert prefix "no" and "inv"
        # to options. For example, "nowrap" is an inverted option "wrap".
        is_valid = option_name not in AbbreviationsIncludingInvertPrefix

        if not is_valid:
            self._make_description_by_option_name(option_name)

        return is_valid


    def _make_description_by_option_name(self, option_name):
        param = {
            'good_pattern': AbbreviationsIncludingInvertPrefix[option_name],
            'bad_pattern': option_name,
        }

        self.description = ('Use the full option name `{good_pattern}` '
                            'instead of `{bad_pattern}`'.format(**param))



from vint.linting.policy.abstract_policy import AbstractPolicy
from vint.linting.policy.reference.googlevimscriptstyleguide import get_reference_source
from vint.linting.level import Level
from vint.ast.node_type import NodeType
from vint.linting.policy_registry import register_policy


GoodStringComparisonOperators = {
    NodeType.EQUAL: ['==#', '==?'],
    NodeType.NEQUAL: ['!=#', '!=?'],
    NodeType.GREATER: ['>#', '>?'],
    NodeType.GEQUAL: ['>=#', '>=?'],
    NodeType.SMALLER: ['<#', '<?'],
    NodeType.SEQUAL: ['<=#', '<=?'],
    NodeType.MATCH: ['=~#', '=~?'],
    NodeType.NOMATCH: ['!~#', '!~?'],
    NodeType.IS: ['is#', 'is?'],
    NodeType.ISNOT: ['isnot#', 'isnot?'],
}


BadStringComparisonOperators = {
    NodeType.EQUAL: '==',
    NodeType.NEQUAL: '!=',
    NodeType.GREATER: '>',
    NodeType.GEQUAL: '>=',
    NodeType.SMALLER: '<',
    NodeType.SEQUAL: '<=',
    NodeType.MATCH: '=~',
    NodeType.NOMATCH: '!~',
    NodeType.IS: 'is',
    NodeType.ISNOT: 'isnot',
}


@register_policy
class ProhibitEqualTildeOperator(AbstractPolicy):
    def __init__(self):
        super(ProhibitEqualTildeOperator, self).__init__()
        self.level = Level.WARNING
        self.description = 'Use the =~# or =~? operator families over the =~ family.'
        self.reference = get_reference_source('MATCHING')


    def listen_node_types(self):
        return [
            NodeType.EQUAL,
            NodeType.NEQUAL,
            NodeType.GREATER,
            NodeType.GEQUAL,
            NodeType.SMALLER,
            NodeType.SEQUAL,
            NodeType.MATCH,
            NodeType.NOMATCH,
            NodeType.IS,
            NodeType.ISNOT,
        ]


    def is_valid(self, node, lint_context):
        """ Whether the specified node is valid to the policy.

        In this policy, comparing between a string and any value by
        'ignorecase'-sensitive is invalid. This policy can detect following
        script: variable =~ '1'

        But detecting exactly string comparison without evaluation is very hard.
        So this policy often get false-positive/negative results.
            False-positive case is: '1' =~ 1
            False-negative case is: ('1') =~ 1
        """
        node_type = NodeType(node['type'])

        left_type = NodeType(node['left']['type'])
        right_type = NodeType(node['right']['type'])

        is_like_string_comparison = left_type is NodeType.STRING \
            or right_type is NodeType.STRING

        is_valid = not is_like_string_comparison

        if not is_valid:
            self._make_description(node_type)

        return is_valid


    def _make_description(self, node_type):
        good_examples = ['`' + op + '`' for op in GoodStringComparisonOperators[node_type]]
        formatted_good_examples = ' or '.join(good_examples)

        bad_example = BadStringComparisonOperators[node_type]

        params = {
            'good_example': formatted_good_examples,
            'bad_example': bad_example,
        }
        self.description = ('Use robust operators {good_example} '
                            'instead of `{bad_example}`').format(**params)



SectionNames = {
    # Portability
    'STRINGS': 'Strings',
    'MATCHING': 'Matching',
    'REGULAR': 'Regular',
    'EXPRESSIONS': 'Expressions',
    'DANGEROUS': 'Dangerous',
    'COMMANDS': 'Commands',
    'FRAGILE': 'Fragile',
    'CATCHING': 'Catching',
    'EXCEPTIONS': 'Exceptions',

    # General Guidelines
    'MESSAGING': 'Messaging',
    'TYPE': 'Type',
    'CHECKING': 'checking',
    'PYTHON': 'Python',
    'OTHER': 'Other',
    'LANGUAGES': 'Languages',
    'BOILERPLATE': 'Boilerplate',
    'PLUGIN': 'Plugin',
    'LAYOUT': 'Layout',
    'FUNCTIONS': 'Functions',
    'AUTOCOMMANDS': 'Autocommands',
    'MAPPINGS': 'Mappings',
    'SETTINGS': 'Settings',

    # Style
    'WHITESPACE': 'Whitespace',
    'NAMING': 'Naming',
}


def get_reference_source(section):
    return 'Google VimScript Style Guide ({section_name})'.format(
        section_name=SectionNames[section])
